import re
import sys
from os.path import isdir, isfile
from os import mkdir


def evaluate_success(control_results, case_results, factor=0.001):
    success = True
    try:
        for key in case_results:
            if case_results[key] >= control_results[key] * factor:
                success = False
    except TypeError:
        success = False
    return success


def hash_from_list(_list):
    _list = list(sorted(_list))
    val = (sys.maxsize + 1) * 2
    return hash(tuple(_list)) + val


def generate_jobscript(template, parameters):
    with open(template) as f:
        job = f.read()
        for parameter, value in parameters.items():
            job = job.replace("%%%s%%" % parameter, value)
    return job


def submit_job(job, remote_jobscript, communicator):
    with communicator.sftp.file(remote_jobscript, "w") as jobscript:
        jobscript.write(job)

    submit_command = "sbatch"
    command = "%s %s" % (submit_command, remote_jobscript)
    stdin, stdout, stderr = communicator.execute(command)
    output = str(stdout.read())

    jobid_pattern = "Submitted batch job +([0-9]+)"
    jobid = re.search(jobid_pattern, output).group(1)
    return jobid


def job_status(job_id, communicator):
    command = 'sacct --job %s --format="State,Elapsed"' % job_id
    stdin, stdout, stderr = communicator.execute(command)
    output = stdout.read().decode()
    try:
        split_output = output.split("\n")[2].split()
        status = split_output[0]
        time = split_output[1]
    except IndexError:
        status = "PENDING"
        time = "00:00"
    return status, time


def obtain_result(remote_rundir, communicator, result_name=None):
        import pickle as pkl
        remote_filename = "%s/%s" % (remote_rundir, result_name)
        # Setting the buffer size to deal with outputs bigger than few kB
        buffer_size = 3 * 1024 * 1024
        with communicator.sftp.open(remote_filename, "rb", bufsize=buffer_size) as f:
            return pkl.load(f)


def generate_namelist(variable_set, filename, _vault, reduced_precision_level=10):
    variables = [v for v in _vault.variables if v.mutable]
    for index, v in enumerate(variables):
        if v.id in variable_set:
            v.precision = reduced_precision_level
        else:
            v.precision = 52
    _vault.generate_namelist_precisions(filename=filename, exclude_double_precision=True)


def var_set_identifier(var_set):
    to_return = ""
    current_var_set = var_set
    while current_var_set.parent is not None:
        if len(to_return) > 100:
            raise AssertionError("Something strange happens here!")
        to_return += str(current_var_set.child_number)
        current_var_set = current_var_set.parent
    to_return = "l%02s:%s" % (str(var_set.level), to_return[::-1])
    return to_return


def divide_set(set0, vault):
    use_old = False
    if use_old:
        return simple_divide_set(set0)
    else:
        return smart_divide(variable_set=set0, vault=vault)


def smart_divide(variable_set, vault):
    variables = [vault.get_variable_by_id(_id) for _id in variable_set]

    # Dividing by modules
    modules = [var.routine.module.name for var in variables]
    # It is necessary to sort the modules otherwise the order can be different in different runs
    # leading to different groups with different hashes, which prevents reproducibility
    unique_modules = list(sorted(set(modules)))
    if len(unique_modules) > 1:
        d = {}
        for mod in unique_modules:
            d[mod] = modules.count(mod)
        d = {k: v for k, v in sorted(d.items(), key=lambda item: - item[1])}
        g1 = {"mod": [], "val": 0}
        g2 = {"mod": [], "val": 0}
        for key, value in d.items():
            if g1["val"] < g2["val"]:
                g1["val"] += value
                g1["mod"].append(key)
            else:
                g2["val"] += value
                g2["mod"].append(key)
        set1 = [v.id for v in variables if v.routine.module.name in g1["mod"]]
        set2 = [v.id for v in variables if v.routine.module.name in g2["mod"]]
        return set1, set2
    else:
        # Separate by routines
        routines = [var.routine.name for var in variables]
        unique_routines = list(sorted(set(routines)))
        if len(unique_routines) > 1:
            d = {}
            for routine in unique_routines:
                d[routine] = routines.count(routine)
            d = {k: v for k, v in sorted(d.items(), key=lambda item: - item[1])}
            g1 = {"mod": [], "val": 0}
            g2 = {"mod": [], "val": 0}
            for key, value in d.items():
                if g1["val"] < g2["val"]:
                    g1["val"] += value
                    g1["mod"].append(key)
                else:
                    g2["val"] += value
                    g2["mod"].append(key)
            set1 = [v.id for v in variables if v.routine.name in g1["mod"]]
            set2 = [v.id for v in variables if v.routine.name in g2["mod"]]
            return set1, set2
        else:
            # Just separate in two halves
            return simple_divide_set(variable_set)


def simple_divide_set(set0):
    assert len(set0) > 1
    set1 = []
    set2 = set0[:]
    for i in range(int(len(set0) / 2)):
        set1.append(set2.pop(0))
    return set1, set2


class Job(object):
    def __init__(self, analysis_set, forced_variables=None, reduced_precision_level=10, communicator=None, vault=None,
                 template="Templates/singleJobscript.sh"):
        # Init variables
        self.vault = vault
        self.analysis_set = analysis_set
        if forced_variables is not None:
            self.forced_variables = forced_variables[:]
        else:
            self.forced_variables = None
        self.communicator = communicator
        self.job_name = "BinarySearch%i" % self.hash
        self.remote_scratch = self.communicator.remote_scratch
        self.remote_logdir = "%s/LOGS" % self.remote_scratch
        self.remote_namelist_dir = "%s/Namelists" % self.remote_logdir

        self.remote_rundir = "%s/Test%i" % (self.remote_scratch, self.hash)

        self.remote_jobscript_path = "%s/test_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/test_namelist_%i" % (self.remote_namelist_dir, self.hash)
        self.local_namelist = "NewWorlds/namelist_precisions"

        self.template = template
        # parameters to replace:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }

        self.jobscript = generate_jobscript(self.template, self.parameters)

        self.elapsed_time = None
        self.job_id = None
        self._result = None

        self.reduced_precision_level = reduced_precision_level

        # The default status is PENDING
        self._status = "PENDING"

    @property
    def variable_set(self):
        if self.forced_variables is not None:
            return self.analysis_set + self.forced_variables
        else:
            return self.analysis_set

    def variables(self):
        return [self.vault.get_variable_by_id(_x) for _x in set(self.variable_set)]

    @property
    def remote_status(self):
        _status, _time = job_status(self.job_id, self.communicator)
        _status = _status.strip()
        return _status

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, new_value):
        possible_status = ["PENDING", "RUNNING", "ASSERTION_PENDING", "SUCCESS", "FAILED", "SUSPENDED"]
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # A PENDING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS
        # RUNNING: The job is RUNNING.
        # A RUNNING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # A SUCCESS job can not have both descendants FAILED
        # FAILED: A job that has been completed with unsuccessful results and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
        # SUSPENDED ancestors must be SUSPENDED.
        # At least one direct descendant should be different from FAILED or COMPLETED
        if new_value in possible_status:
            self._status = new_value
        else:
            raise AssertionError("Unknown status")

    def result(self):
        # This method will be over-written in BinarySearchJob
        #self._result = obtain_result(self.remote_rundir, self.communicator, result_name=self.analysis_set.result_name)
        return self._result

    @property
    def hash(self):
        return hash_from_list(self.variable_set)

    def recover_job_id(self):
        job_id_filename = "%s/job_id.txt" % self.remote_rundir
        self.job_id = self.communicator.sftp.open(job_id_filename).read().decode().strip()

    def check_if_result_already_exists(self):
        self.update_parameters()
        # TODO: find a way to extract accuracy test from the core code. The way of testing a result should be provided
        # as an argument.
        filename = "simulation_RMSE.pkl"
        accuracy_filename = "%s/%s" % (self.remote_rundir, filename)
        try:
            self.communicator.sftp.open(accuracy_filename)
            # Recovering job_id
            self.recover_job_id()
            return True
        except IOError:
            return False

    def check_if_already_submitted(self):
        self.update_parameters()
        try:
            command = 'squeue --format="%i %j"'
            stdin, stdout, stderr = self.communicator.execute(command)

            output = stdout.read().decode()
            lines = output.split("\n")
            if len(lines) > 1:
                jobs_in_remote_queue = lines[1:]
                for line in jobs_in_remote_queue:
                    if line.strip():
                        job_id, job_name = line.split()
                        if job_name == self.job_name:
                            return job_id

        except IOError:
            return None
        return None

    def submit(self, force=False):
        # Check if result already exists
        if force is False:
            completed = self.check_if_result_already_exists()
        else:
            completed = False
        if not completed:
            self.update_parameters()
            # Generate namelist and put it in the remote system
            generate_namelist(self.variable_set, self.local_namelist, self.vault, self.reduced_precision_level)

            self.communicator.put(self.local_namelist, self.remote_namelist)
            # Submit job and obtain job id
            if force is False:
                self.job_id = self.check_if_already_submitted()
            else:
                self.job_id = None
            if self.job_id is None:
                self.job_id = submit_job(self.jobscript, self.remote_jobscript_path, self.communicator)
                print("Job %i successfully submitted to the remote machine with job_id %s" % (self.hash, self.job_id))
            else:
                print("Job %i was already in queue with job_id %s" % (self.hash, self.job_id))
        else:
            pass

    def update_parameters(self):
        self.remote_rundir = "%s/Test%i" % (self.remote_scratch, self.hash)
        self.job_name = "BinarySearch%i" % self.hash

        self.remote_jobscript_path = "%s/test_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/test_namelist_%i" % (self.remote_namelist_dir, self.hash)
        namelist_folder = "PrecisionNamelists"
        # TODO: Put it in a better place
        if not isdir(namelist_folder):
            mkdir(namelist_folder)
        self.local_namelist = "%s/namelist_precisions_%i" % (namelist_folder, self.hash)

        # Set of parameters that will be replaced in the job script:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }
        self.jobscript = generate_jobscript(self.template, self.parameters)

    def compare_with_job(self, other_job, outputs_file_name="", script=""):
        folder_files_2 = other_job.communicator.list_dir(other_job.remote_rundir)
        filename_2 = [f for f in folder_files_2 if re.search(outputs_file_name, f)][0]
        results_2 = "%s/%s" % (other_job.remote_rundir, filename_2)

        return self.compare_with_file(results_2, outputs_file_name, script)

    def compare_with_file(self, file_path, outputs_file_name="", script=""):
        folder_files_1 = self.communicator.list_dir(self.remote_rundir)
        filename_1 = [f for f in folder_files_1 if re.search(outputs_file_name, f)][0]
        results_1 = "%s/%s" % (self.remote_rundir, filename_1)

        results_2 = file_path

        with open(script) as input_script:
            remote_script = input_script.read()

        remote_script = remote_script.replace("%FILENAME1%", results_1)
        remote_script = remote_script.replace("%FILENAME2%", results_2)

        local_wrapper_path = "/home/Earth/otintopr/RPE-Canal/AutoRPE/EvaluationScripts/wrapper.sh"
        remote_wrapper_path = "%s/python_wrapper.sh" % self.remote_logdir

        self.communicator.put(local_wrapper_path, remote_wrapper_path)

        remote_script_name = "compare_script.py"
        remote_script_path = "%s/%s" % (self.remote_logdir, remote_script_name)
        with self.communicator.sftp.file(remote_script_path, "w") as remote_script_file:
            remote_script_file.write(remote_script)

        # Change permissions for the wrapper
        command = "chmod 777 %s" % remote_wrapper_path
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read())

        # Execute script through wrapper
        command = "%s %s" % (remote_wrapper_path, remote_script_path)
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read().decode())
        err = str(stderr.read().decode())

        if output.count("not found!"):
            print("File not found")
            return None

        output = [line for line in output.split("\n") if line.strip()]
        results = {}
        for line in output:
            name, value = line.split()
            value = float(value)
            results[name] = value

        if err.strip():
            print(err)
        return results


class BinarySearchJob(Job):
    def __init__(self,
                 analysis_set,
                 forced_variables=None,
                 reduced_precision_level=10,
                 communicator=None,
                 vault=None,
                 template="",
                 limit_file=None,
                 local_folder=None,
                 accuracy_test=None,
                 ):
        super().__init__(analysis_set,
                         forced_variables=forced_variables,
                         reduced_precision_level=reduced_precision_level,
                         communicator=communicator,
                         vault=vault,
                         template=template)
        # This class extends the classical job putting some extra variables that define the relations
        self.parent = None
        self.child = None
        self.child_number = None
        self.level = None
        self.banned_variables = []
        self.thresholds = {}
        self.limit_file = limit_file
        self.local_folder = local_folder
        self.max_retrials = 1
        self.retrials = 0
        self.accuracy_test = accuracy_test

    def identifier(self):
        return var_set_identifier(self)

    # The following funtions have been extracted from this code and are now provided as an Evaluator class, which
    # is stored in self.accuracy_test, and which has the evaluate_success, evaluate_fitness and plot methods that are
    # used here.
    def evaluate_success(self):
        return self.accuracy_test.evaluate_success(self.result())

    def evaluate_fitness(self):
        return self.accuracy_test.evaluate_fitness(self.result())

    def create_plots(self, force=False):
        return self.accuracy_test.plot(self.result())
        # from os.path import join, isdir
        # from os import makedirs
        # from AutoRPE.EvaluationScripts.plot_result import plot_result
        # from AutoRPE.EvaluationScripts.variable_summary import variable_summary
        # from AutoRPE.EvaluationScripts.accuracy_test_with_limit_file import list_of_failed_variables
        # from AutoRPE.UtilsRPE.Functions import save_variable_list
        # import pickle as pkl
        #
        # # Setting path to generate the plots
        # path_in_folder = "Plots/%s/%s" % (self.identifier(), self.hash)
        # # Removing spaces and colons for clarity
        # path_in_folder = path_in_folder.replace(" ", "")
        # path_in_folder = path_in_folder.replace(":", "_")
        # destination_folder = join(self.local_folder, path_in_folder)
        #
        # # Check that output folder exists and otherwise create folder
        # if isdir(destination_folder):
        #     if not force:
        #         print("%s already exists" % destination_folder)
        #         return list_of_failed_variables(self.result(accuracy_file=False), self.limit_file), destination_folder
        # else:
        #     makedirs(destination_folder)
        #
        # self.result(accuracy_file=False)
        #
        # # Create summary to add a description of the experiment variables in the viewer.
        # summary = variable_summary(self.variables())
        # plot_result(self._result, destination_folder, limits=self.limit_file, variable_summary=summary)
        #
        # # Save list of analysis variables
        # analysis_variables_filename = join(destination_folder, "analysis_variables.txt")
        # save_variable_list(self.analysis_set, analysis_variables_filename, self.vault)
        #
        # # Save list of forced variables
        # if self.forced_variables is not None:
        #     forced_variables_filename = join(destination_folder, "forced_variables.txt")
        #     save_variable_list(self.forced_variables, forced_variables_filename, self.vault)
        #
        # # Save RMSE file.
        # rmse_filename = join(destination_folder, "result.pkl")
        # pkl.dump(self._result, open(rmse_filename, "wb"))
        #
        # return list_of_failed_variables(self._result, self.limit_file), destination_folder

    def result(self):
        if self._result is None:
            result_filename = self.accuracy_test.result_filename
            self._result = obtain_result(self.remote_rundir, self.communicator, result_name=result_filename)
        return self._result

    def ancestors(self):
        to_return = []
        cv = self
        while cv.parent is not None:
            to_return.append(cv.parent)
            cv = cv.parent
        # return list(reversed(to_return))
        return list(to_return)

    def descendants(self):
        if self.child is None:
            return []
        to_return = []
        pre_list = [child for child in self.child]
        while len(pre_list) > 0:
            cv = pre_list.pop(0)
            if cv.child is not None:
                for child in cv.child:
                    if child not in pre_list:
                        pre_list.append(child)
                to_return.append(cv)
        return to_return

    def spawn_children(self, queue):
        # In case the set has not been divided already
        if self.child is None:

            # Trying new thing, set the minimum unit to explore to modules
            # Check if all variables are from the same module
            variables = [self.vault.get_variable_by_id(_id) for _id in self.analysis_set]
            routines = set([v.routine for v in variables])
            modules = set([r.module for r in routines])
            if len(modules) == 1:
                return False
            if len(self.analysis_set) >= 2:
                if self.thresholds is not None:
                    child_thresholds = dict(self.thresholds)
                    for key in child_thresholds:
                        factor = (3. / 4.)
                        child_thresholds[key] = child_thresholds[key] * factor
                    child_thresholds = dict(child_thresholds)
                else:
                    child_thresholds = None
                _set1, _set2 = divide_set(self.analysis_set, self.vault)
                self.child = []
                for index, var_set in enumerate([_set1, _set2]):
                    job = BinarySearchJob(var_set,
                                          forced_variables=self.forced_variables,
                                          reduced_precision_level=self.reduced_precision_level,
                                          communicator=self.communicator,
                                          vault=self.vault,
                                          template=self.template,
                                          accuracy_test=self.accuracy_test,
                                          local_folder=self.local_folder,
                                          )
                    job.parent = self
                    job.child_number = index
                    job.level = self.level + 1
                    job.thresholds = child_thresholds
                    self.child.append(job)
                return True
            else:
                return False
        else:
            return False

    def driver(self, queue, max_running_jobs):
        status = self.status
        if self.status == "PENDING":
            if len(queue.running) < max_running_jobs:
                self.submit()
                self.status = "RUNNING"
                queue.pending.remove(self)
                queue.running.append(self)
        elif self.status == "RUNNING":
            self.check_running(queue)
        elif self.status == "SUSPENDED":
            self.check_child(queue)

        # Return True if there was a status change, False otherwise
        return status != self.status

    def check_running(self, queue):
        remote_status = self.remote_status
        if remote_status == "COMPLETED":
            self.retrials = 0
            success = self.evaluate_success()
            # Create plots
            # self.create_plots()
            if success:
                self.qualify(queue)
            else:
                self.fail(queue)

        elif remote_status == "FAILED":
            if self.retrials < self.max_retrials:
                self.submit(force=True)
                self.retrials += 1
            else:
                raise AssertionError("SLURM job failed!")
            # self.fail(queue)
        elif remote_status == 'CANCELLED+':
            self.submit(force=True)
            self.retrials = 0
        elif remote_status == "RUNNING":
            pass
        elif remote_status == "PENDING":
            pass
        elif remote_status == "TIMEOUT":
            self.submit(force=True)
        else:
            raise AssertionError("Unhandled status %s" % remote_status)

    def move_to_success(self, queue):
        queue[self.status].remove(self)
        self.status = "SUCCESS"
        queue.success.append(self)

    def qualify(self, queue):
        self.move_to_success(queue)

    def fail(self, queue):

        # If the set can be split
        if self.spawn_children(queue):
            print("Job %s %s suspended with fitness: %.2f." % (self.identifier(), self.hash, self.evaluate_fitness()))
            queue[self.status].remove(self)
            queue.suspended.append(self)
            self.status = "SUSPENDED"
            for ch in self.child:
                if ch not in queue.all():
                    queue.pending.append(ch)
        # If the set can not be split, check whether its a exception and handle it
        else:
            if self.kind_of_exception() == NoException:
                queue[self.status].remove(self)
                self.status = "FAILED"
                queue.failed.append(self)
                print("Job %s %s failed with fitness: %.2f." % (self.identifier(), self.hash, self.evaluate_fitness()))
            else:
                # Exception handling
                print(
                    "Job %s %s suspended with fitness: %.2f." % (self.identifier(), self.hash, self.evaluate_fitness()))
                print("Exception found!", self.kind_of_exception())
                queue[self.status].remove(self)
                queue.suspended.append(self)
                self.status = "SUSPENDED"

                child_1, child_2 = self.child
                # Make that the analysis variables of one child are the forced variables of the other one
                if child_1.forced_variables is not None:
                    child_1.forced_variables.extend(child_2.analysis_set)
                    child_2.forced_variables.extend(child_1.analysis_set)
                else:
                    child_1.forced_variables = child_2.analysis_set
                    child_2.forced_variables = child_1.analysis_set
                for ch in self.child:
                    if ch in queue.all():
                        queue[ch.status].remove(ch)
                    queue.pending.append(ch)
                    ch.status = "PENDING"
                    # Disinherit all child's descendants
                    for job in ch.descendants():
                        job.parent = None
                    ch.child = None
                # Save disinherited jobs
                disinherited = [j for j in queue.all() if j.parent is None and j.level != 0]
                for job in disinherited:
                    queue[job.status].remove(job)
                    if job not in queue.disinherited:
                        queue.disinherited.append(job)

    def check_child(self, queue):
        statuses = list(sorted([c.status for c in self.child]))
        if statuses == ["FAILED", "SUCCESS"]:
            pass

        all_finished = all([c.status == "FAILED" or c.status == "SUCCESS" for c in self.child])
        if all_finished:
            for child in self.child:
                self.banned_variables += child.banned_variables
                if child.status == "FAILED":
                    self.banned_variables += child.analysis_set
            self.analysis_set = [var for var in self.analysis_set if var not in self.banned_variables]

            if self.analysis_set is []:
                self.fail(queue)
            else:
                self.status = "PENDING"
                queue.pending.append(self)
                queue.suspended.remove(self)

    def kind_of_exception(self):
        """
        This method tries to categorize the kind of exception that occurs when two successful sets merge to an
        unsuccessful one.

        To start with something I propose:
        - No exception: It isn't a failed job
        - Intra-routine: All variables belong to same routine
        - Intra-module: All variables belong to same module, but not routine
        - Inter-module: The variables belong to different modules.
        :return: kind-of-exception
        """
        if not self.child:
            return NoException
        child_1, child_2 = self.child
        # We'll focus only on the analysis_set
        ch1_variables = [self.vault.get_variable_by_id(_id) for _id in child_1.analysis_set]
        ch2_variables = [self.vault.get_variable_by_id(_id) for _id in child_2.analysis_set]
        ch1_routines = set([v.routine for v in ch1_variables])
        ch2_routines = set([v.routine for v in ch2_variables])
        if len(ch1_routines.union(ch2_routines)) == 1:
            return IntraRoutine
        ch1_modules = set([v.routine.module for v in ch1_variables])
        ch2_modules = set([v.routine.module for v in ch2_variables])
        if len(ch1_modules.union(ch2_modules)) == 1:
            return IntraModule
        return InterModule




class ExperimentDriver(object):
    def __init__(self, seed_job=None, max_running_jobs=100):
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # RUNNING: The job is RUNNING.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # FAILED: A job that has been completed with unsuccessful results and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
        self.loop = 0
        self.pending = []
        self.running = []
        self.success = []
        self.failed = []
        self.suspended = []
        self.disinherited = []
        self.__dict = {"PENDING": self.pending,
                       "RUNNING": self.running,
                       "SUCCESS": self.success,
                       "FAILED": self.failed,
                       "SUSPENDED": self.suspended,
                       "DISINHERITED": self.disinherited,
                       }
        self.seed_job = seed_job
        self.max_running_jobs = max_running_jobs
        if self.seed_job is not None:
            self.pending.append(self.seed_job)
            self.checkpoint_name = "checkpoint_%i.pkl" % self.__hash__()
        else:
            self.checkpoint_name = "checkpoint.pkl"

    def sort(self, lower_first=True):
        if lower_first:
            self.pending.sort(key=lambda _x: _x.level)
        else:
            self.pending.sort(key=lambda _x: - _x.level)

    def all(self):
        return self.running + self.pending + self.success + self.failed + self.suspended

    def __getitem__(self, item):
        return self.__dict[item]

    def step(self):
        """
        The function will loop through the jobs in the queue and will execute the driver.
        :return: Number of jobs that changed status
        """
        changes = 0
        for job in self.all():
            if job.driver(queue=self, max_running_jobs=self.max_running_jobs): changes += 1
        self.loop += changes
        return changes

    def print_status(self):
        for key in self.__dict.keys():
            print("\t%18s: %4i" % (key, len(self[key])))

    def run(self, condition):
        import time

        loops = 0
        change_counter = 0
        wait = False
        waiting_time = 10
        while condition(self.seed_job):
            # Trying to avoid too much ssh polling by waiting if the loop time is less than a threshold
            loop_start_time = time.time()
            # Main loop
            print("ITERATION: %4i" % loops)
            self.print_status()
            # Sorting the queue to launch first the lower levels (bigger groups).
            self.sort(lower_first=True)
            # Launch one step (will loop through the jobs in the queue and take actions according to their status)
            change_counter += self.step()
            if change_counter > 10:
                self.checkpoint()
                change_counter = 0
            loop_end_time = time.time()
            if wait:
                while loop_end_time - loop_start_time < waiting_time:
                    time.sleep(1)
                    loop_end_time = time.time()
        print("done")

    def checkpoint(self, filename=None):
        from AutoRPE.UtilsWorkflow.CheckpointTools import checkpoint
        if filename is None:
            filename = self.checkpoint_name
        checkpoint(queue=self, iteration=self.loop, filename=filename)

    def __hash__(self):
        return self.seed_job.hash


# Exceptions
class WorkflowException(Exception):
    """Base class for other exceptions"""
    pass


class NoException(WorkflowException):
    """No exception is found"""
    pass


class IntraRoutine(WorkflowException):
    """All variables belong to same routine"""
    pass


class IntraModule(WorkflowException):
    """All variables belong to same module, but no same routine"""


class InterModule(WorkflowException):
    """Variables belong to different modules"""
